"""Tests for `ember_mug.mug connections`."""
from __future__ import annotations

from unittest.mock import AsyncMock, Mock, patch

import pytest
from bleak import BleakError
from bleak.backends.device import BLEDevice

from ember_mug.consts import MugCharacteristic, TemperatureUnit
from ember_mug.data import Colour
from ember_mug.mug import EXTRA_ATTRS, INITIAL_ATTRS, UPDATE_ATTRS, EmberMug


@patch('ember_mug.mug.IS_LINUX', True)
def test_adapter_with_bluez(ble_device: BLEDevice):
    mug = EmberMug(ble_device, adapter='hci0')
    assert mug._client_kwargs['adapter'] == 'hci0'


@patch('ember_mug.mug.IS_LINUX', False)
def test_adapter_without_bluez(ble_device: BLEDevice):
    with pytest.raises(ValueError):
        EmberMug(ble_device, adapter='hci0')


@patch('ember_mug.mug.EmberMug.subscribe')
@patch('ember_mug.mug.establish_connection')
async def test_connect(
    mug_subscribe: AsyncMock,
    mock_establish_connection: AsyncMock,
    ember_mug: AsyncMock,
) -> None:
    # Already connected
    ember_mug._client = AsyncMock()
    ember_mug._client.is_connected = True
    async with ember_mug.connection():
        pass
    mug_subscribe.assert_not_called()
    mock_establish_connection.assert_not_called()

    # Not connected
    ember_mug._client = None
    ember_mug.disconnect = AsyncMock()
    async with ember_mug.connection():
        pass

    mock_establish_connection.assert_called()
    mug_subscribe.assert_called()
    assert ember_mug._client is not None
    ember_mug.disconnect.assert_called()


@patch('ember_mug.mug.logger')
@patch('ember_mug.mug.establish_connection')
async def test_connect_error(
    mock_establish_connection: AsyncMock,
    mock_logger: Mock,
    ember_mug: AsyncMock,
) -> None:
    ember_mug._client = None  # type: ignore[assignment]
    mock_establish_connection.side_effect = BleakError
    with pytest.raises(BleakError):
        await ember_mug._ensure_connection()
    msg, device, exception = mock_logger.error.mock_calls[0].args
    assert msg == "%s: Failed to connect to the mug: %s"
    assert device == ember_mug.device
    assert isinstance(exception, BleakError)


@patch('ember_mug.mug.logger')
@patch('ember_mug.mug.establish_connection')
async def test_pairing_exceptions_esphome(
    mock_establish_connection: AsyncMock,
    mock_logger: Mock,
    ember_mug: AsyncMock,
) -> None:
    mock_client = AsyncMock()
    mock_client.connect.side_effect = BleakError
    mock_client.pair.side_effect = NotImplementedError
    mock_establish_connection.return_value = mock_client
    with patch.multiple(
        ember_mug,
        update_initial=AsyncMock(),
        subscribe=AsyncMock(),
    ):
        await ember_mug._ensure_connection()

    mock_establish_connection.assert_called_once()
    mock_logger.warning.assert_called_with(
        'Pairing not implemented. '
        'If your mug is still in pairing mode (blinking blue) tap the button on the bottom to exit.',
    )


@patch('ember_mug.mug.establish_connection')
async def test_pairing_exceptions(
    mock_establish_connection: AsyncMock,
    ember_mug: AsyncMock,
) -> None:
    mock_client = AsyncMock()
    mock_client.pair.side_effect = BleakError
    mock_establish_connection.return_value = mock_client
    with patch.multiple(
        ember_mug,
        update_initial=AsyncMock(),
        subscribe=AsyncMock(),
    ):
        await ember_mug._ensure_connection()


async def test_disconnect(ember_mug: AsyncMock) -> None:
    ember_mug._client = AsyncMock()

    ember_mug._client.is_connected = False
    await ember_mug.disconnect()
    ember_mug._client.disconnect.assert_not_called()

    ember_mug._client.is_connected = True
    await ember_mug.disconnect()
    ember_mug._client.disconnect.assert_called()


@patch('ember_mug.mug.logger')
def test_disconnect_callback(mock_logger: Mock, ember_mug: AsyncMock) -> None:
    ember_mug._expected_disconnect = True
    ember_mug._disconnect_callback(AsyncMock())
    mock_logger.debug.assert_called_with("Disconnect callback called")
    mock_logger.reset_mock()

    ember_mug._expected_disconnect = False
    ember_mug._disconnect_callback(AsyncMock())
    mock_logger.warning.assert_called_with("Unexpectedly disconnected")


@patch('ember_mug.mug.logger')
async def test_read(
    mock_logger: Mock,
    ember_mug: AsyncMock,
) -> None:
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'TEST'
        await ember_mug._read(MugCharacteristic.MUG_NAME)
        ember_mug._client.read_gatt_char.assert_called_with(
            MugCharacteristic.MUG_NAME.uuid,
        )
        mock_logger.debug.assert_called_with(
            "Read attribute '%s' with value '%s'",
            MugCharacteristic.MUG_NAME,
            b'TEST',
        )


@patch('ember_mug.mug.logger')
async def test_write(mock_logger: Mock, ember_mug: AsyncMock) -> None:
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        test_name = bytearray(b'TEST')
        await ember_mug._write(
            MugCharacteristic.MUG_NAME,
            test_name,
        )
        ember_mug._client.write_gatt_char.assert_called_with(
            MugCharacteristic.MUG_NAME.uuid,
            test_name,
        )
        mock_logger.debug.assert_called_with(
            "Wrote '%s' to attribute '%s'",
            test_name,
            MugCharacteristic.MUG_NAME,
        )

        ember_mug._client = AsyncMock()
        ember_mug._client.write_gatt_char.side_effect = BleakError
        with pytest.raises(BleakError):
            await ember_mug._write(
                MugCharacteristic.MUG_NAME,
                test_name,
            )
        ember_mug._client.write_gatt_char.assert_called_with(
            MugCharacteristic.MUG_NAME.uuid,
            test_name,
        )
        msg, data, char, exception = mock_logger.error.mock_calls[0].args
        assert msg == "Failed to write '%s' to attribute '%s': %s"
        assert data == test_name
        assert char == MugCharacteristic.MUG_NAME
        assert isinstance(exception, BleakError)


def test_set_device(ember_mug: AsyncMock) -> None:
    new_device = BLEDevice(
        address='BA:36:a5:be:88:cb',
        name='Ember Ceramic Mug',
    )
    assert ember_mug.device.address != new_device.address
    ember_mug.set_device(new_device)
    assert ember_mug.device.address == new_device.address


async def test_get_mug_meta(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'Yw====-ABCDEFGHIJ'
        meta = await ember_mug.get_meta()
        assert meta.mug_id == 'WXc9PT09'
        assert meta.serial_number == 'ABCDEFGHIJ'
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.MUG_ID.uuid)


async def test_get_mug_battery(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'5\x01'
        battery = await ember_mug.get_battery()
        assert battery.percent == 53.00
        assert battery.on_charging_base is True
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.BATTERY.uuid)


async def test_get_mug_led_colour(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'\xf4\x00\xa1\xff'
        colour = await ember_mug.get_led_colour()
        assert colour.as_hex() == '#f400a1'
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.LED.uuid)


async def test_set_mug_led_colour(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        await ember_mug.set_led_colour(Colour(244, 0, 161))
        ember_mug._ensure_connection.assert_called_once()
        ember_mug._client.write_gatt_char.assert_called_once_with(
            MugCharacteristic.LED.uuid,
            bytearray(b'\xf4\x00\xa1\xff'),
        )


async def test_get_mug_target_temp(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'\xcd\x15'
        assert (await ember_mug.get_target_temp()) == 55.81
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.TARGET_TEMPERATURE.uuid)


async def test_set_mug_target_temp(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        await ember_mug.set_target_temp(55.81)
        ember_mug._ensure_connection.assert_called_once()
        ember_mug._client.write_gatt_char.assert_called_once_with(
            MugCharacteristic.TARGET_TEMPERATURE.uuid,
            bytearray(b'\xcd\x15'),
        )


async def test_get_mug_current_temp(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'\xcd\x15'
        assert (await ember_mug.get_current_temp()) == 55.81
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.CURRENT_TEMPERATURE.uuid)


async def test_get_mug_liquid_level(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'\n'
        assert (await ember_mug.get_liquid_level()) == 10
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.LIQUID_LEVEL.uuid)


async def test_get_mug_liquid_state(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'\x06'
        assert (await ember_mug.get_liquid_state()) == 6
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.LIQUID_STATE.uuid)


async def test_get_mug_name(ember_mug):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'Mug Name'
        assert (await ember_mug.get_name()) == 'Mug Name'
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.MUG_NAME.uuid)


async def test_set_mug_name(ember_mug):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        with pytest.raises(ValueError):
            await ember_mug.set_name('Hé!')

    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        await ember_mug.set_name('Mug name')
        ember_mug._ensure_connection.assert_called()
        ember_mug._client.write_gatt_char.assert_called_once_with(
            MugCharacteristic.MUG_NAME.uuid,
            bytearray(b'Mug name'),
        )


async def test_get_mug_udsk(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'abcd12345'
        assert (await ember_mug.get_udsk()) == 'YWJjZDEyMzQ1'
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.UDSK.uuid)


async def test_set_mug_udsk(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        await ember_mug.set_udsk('abcd12345')
        ember_mug._ensure_connection.assert_called_once()
        ember_mug._client.write_gatt_char.assert_called_once_with(
            MugCharacteristic.UDSK.uuid,
            bytearray(b'YWJjZDEyMzQ1'),
        )


async def test_get_mug_dsk(ember_mug):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'abcd12345'
        assert (await ember_mug.get_dsk()) == 'YWJjZDEyMzQ1'
        ember_mug._client.read_gatt_char.return_value = b'something else'
        assert (await ember_mug.get_dsk()) == "c29tZXRoaW5nIGVsc2U="
        ember_mug._client.read_gatt_char.assert_called_with(MugCharacteristic.DSK.uuid)


async def test_get_mug_temperature_unit(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'\x01'
        assert (await ember_mug.get_temperature_unit()) == TemperatureUnit.FAHRENHEIT
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.TEMPERATURE_UNIT.uuid)
        ember_mug._client.read_gatt_char.reset_mock()
        ember_mug._client.read_gatt_char.return_value = b'\x00'
        assert (await ember_mug.get_temperature_unit()) == TemperatureUnit.CELSIUS
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.TEMPERATURE_UNIT.uuid)


async def test_set_mug_temperature_unit(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        await ember_mug.set_temperature_unit(TemperatureUnit.CELSIUS)
        ember_mug._ensure_connection.assert_called_once()
        ember_mug._client.write_gatt_char.assert_called_once_with(
            MugCharacteristic.TEMPERATURE_UNIT.uuid,
            bytearray(b'\x00'),
        )


async def test_mug_ensure_correct_unit(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug.data.temperature_unit = TemperatureUnit.CELSIUS
        ember_mug.data.use_metric = True
        ember_mug.set_temperature_unit = AsyncMock(return_value=None)
        await ember_mug.ensure_correct_unit()
        ember_mug.set_temperature_unit.assert_not_called()
        ember_mug.data.temperature_unit = TemperatureUnit.FAHRENHEIT
        await ember_mug.ensure_correct_unit()
        ember_mug.set_temperature_unit.assert_called_with(TemperatureUnit.CELSIUS)


async def test_get_mug_battery_voltage(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'\x01'
        assert (await ember_mug.get_battery_voltage()) == 1
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.CONTROL_REGISTER_DATA.uuid)


async def test_get_mug_date_time_zone(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'c\x0f\xf6\x00'
        date_time = await ember_mug.get_date_time_zone()
        assert date_time.timestamp() == 1661990400.0
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.DATE_TIME_AND_ZONE.uuid)


async def test_read_firmware(ember_mug: AsyncMock):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._client.read_gatt_char.return_value = b'c\x01\x80\x00\x12\x00'
        firmware = await ember_mug.get_firmware()
        assert firmware.version == 355
        assert firmware.hardware == 128
        assert firmware.bootloader == 18
        ember_mug._client.read_gatt_char.assert_called_once_with(MugCharacteristic.FIRMWARE.uuid)


async def test_mug_update_initial(ember_mug):
    no_extra = INITIAL_ATTRS - EXTRA_ATTRS

    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._update_multiple = AsyncMock(return_value={})
        assert (await ember_mug.update_initial()) == {}
        ember_mug._update_multiple.assert_called_once_with(no_extra)

    # Try with extra
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._update_multiple.reset_mock()
        ember_mug._initial_attrs = INITIAL_ATTRS
        assert (await ember_mug.update_initial()) == {}
        ember_mug._update_multiple.assert_called_once_with(INITIAL_ATTRS)


async def test_mug_update_all(ember_mug):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._update_multiple = AsyncMock(return_value={})
        assert (await ember_mug.update_all()) == {}
        ember_mug._update_multiple.assert_called_once_with(UPDATE_ATTRS - EXTRA_ATTRS)

    # Try with extras
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._update_multiple.reset_mock()
        ember_mug._update_attrs = UPDATE_ATTRS
        assert (await ember_mug.update_all()) == {}
        ember_mug._update_multiple.assert_called_once_with(UPDATE_ATTRS)


async def test_mug_update_multiple(ember_mug):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug.get_name = AsyncMock(return_value='name')
        ember_mug.data.update_info = AsyncMock()
        await ember_mug._update_multiple(('name',))
        ember_mug.data.update_info.assert_called_once_with(name='name')


async def test_mug_update_queued_attributes(ember_mug):
    with patch.object(ember_mug, '_ensure_connection', AsyncMock()):
        ember_mug._queued_updates = set()
        assert (await ember_mug.update_queued_attributes()) == []
        ember_mug.get_name = AsyncMock(return_value='name')
        ember_mug.data.update_info = AsyncMock()
        ember_mug._queued_updates = {'name'}
        await ember_mug.update_queued_attributes()
        ember_mug.data.update_info.assert_called_once_with(name='name')


def test_mug_notify_callback(ember_mug: EmberMug) -> None:
    gatt_char = AsyncMock()
    ember_mug._notify_callback(gatt_char, bytearray(b'\x01'))
    ember_mug._notify_callback(gatt_char, bytearray(b'\x02'))
    assert 2 in ember_mug._latest_events
    ember_mug._notify_callback(gatt_char, bytearray(b'\x04'))
    assert 4 in ember_mug._latest_events
    ember_mug._notify_callback(gatt_char, bytearray(b'\x05'))
    assert 5 in ember_mug._latest_events
    ember_mug._notify_callback(gatt_char, bytearray(b'\x06'))
    assert 6 in ember_mug._latest_events
    ember_mug._notify_callback(gatt_char, bytearray(b'\x07'))
    assert 7 in ember_mug._latest_events
    ember_mug._notify_callback(gatt_char, bytearray(b'\x08'))
    assert 8 in ember_mug._latest_events
    callback = Mock()
    second_callback = Mock()
    unregister = ember_mug.register_callback(callback)
    second_unregister = ember_mug.register_callback(second_callback)
    repeat_unregister = ember_mug.register_callback(callback)
    assert unregister is repeat_unregister
    assert unregister is not second_unregister

    assert callback in ember_mug._callbacks
    ember_mug._notify_callback(gatt_char, bytearray(b'\x09'))
    assert 9 in ember_mug._latest_events
    callback.assert_not_called()
    assert ember_mug._queued_updates == {
        "battery",
        "target_temp",
        "current_temp",
        "liquid_level",
        "liquid_state",
        "battery_voltage",
    }
    ember_mug._latest_events = {}
    ember_mug._notify_callback(gatt_char, bytearray(b'\x02'))
    callback.assert_called_once()
    callback.reset_mock()
    ember_mug._notify_callback(gatt_char, bytearray(b'\x02'))
    callback.assert_not_called()
    # Remove callback
    unregister()
    assert callback not in ember_mug._callbacks



from __future__ import annotations

from collections.abc import AsyncGenerator, Generator
from unittest.mock import AsyncMock, MagicMock

import pytest
import pytest_asyncio
from bleak.backends.device import BLEDevice

from ember_mug import EmberMug
from ember_mug.data import MugData

TEST_MAC = '32:36:a5:be:88:cb'
TEST_MODEL_NAME = 'Ember Ceramic Mug'


class AsyncContextManager:
    """Stub for mocking."""

    async def __aenter__(self):
        """Return self."""
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Close and do nothing."""
        pass


mock_connection = MagicMock(AsyncContextManager)


@pytest.fixture(name='ble_device')
def ble_device_fixture() -> Generator[BLEDevice, None, None]:
    yield BLEDevice(address=TEST_MAC, name=TEST_MODEL_NAME)


@pytest.fixture
def mug_data(ble_device: BLEDevice) -> Generator[MugData, None, None]:
    yield MugData(ble_device.name)


@pytest_asyncio.fixture
async def ember_mug(ble_device: BLEDevice) -> AsyncGenerator[EmberMug | AsyncMock, None]:
    mug = EmberMug(ble_device)
    mug._client = AsyncMock()
    yield mug



"""Objects and methods related to connection to the mug."""
from __future__ import annotations

import asyncio
import contextlib
import logging
from collections.abc import AsyncIterator
from datetime import datetime, timezone
from enum import Enum
from time import time
from typing import Any, Callable, Literal

from bleak import BleakClient, BleakError
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak_retry_connector import IS_LINUX, close_stale_connections, establish_connection

from .consts import (
    EXTRA_ATTRS,
    INITIAL_ATTRS,
    MUG_NAME_REGEX,
    PUSH_EVENT_BATTERY_IDS,
    UPDATE_ATTRS,
    LiquidState,
    MugCharacteristic,
    PushEvent,
    TemperatureUnit,
)
from .data import BatteryInfo, Change, Colour, MugData, MugFirmwareInfo, MugMeta
from .utils import bytes_to_big_int, bytes_to_little_int, decode_byte_string, encode_byte_string, temp_from_bytes

logger = logging.getLogger(__name__)

DEFAULT_ATTEMPTS = 3


class EmberMug:
    """Handle actual the actual mug connection and update states."""

    def __init__(
        self,
        ble_device: BLEDevice,
        include_extra: bool = False,
        use_metric: bool = True,
        **kwargs: Any,
    ) -> None:
        """Initialize connection manager."""
        self.device = ble_device
        self.data = MugData(ble_device.name, include_extra=include_extra, use_metric=use_metric)

        self._connect_lock = asyncio.Lock()
        self._operation_lock = asyncio.Lock()
        self._expected_disconnect = False
        self._callbacks: dict[Callable[[MugData], None], Callable[[], None]] = {}
        self._client: BleakClient = None  # type: ignore[assignment]
        self._queued_updates: set[str] = set()
        self._latest_events: dict[int, float] = {}
        self._client_kwargs: dict[str, str] = {}
        self._initial_attrs = INITIAL_ATTRS if include_extra else (INITIAL_ATTRS - EXTRA_ATTRS)
        self._update_attrs = UPDATE_ATTRS if include_extra else (UPDATE_ATTRS - EXTRA_ATTRS)

        logger.debug("New mug connection initialized.")
        self.set_client_options(**kwargs)

    def set_device(self, ble_device: BLEDevice) -> None:
        """Set the ble device."""
        logger.debug("Set new device from %s to %s", self.device, ble_device)
        self.device = ble_device

    async def _ensure_connection(self) -> None:
        """Connect to mug."""
        if self._connect_lock.locked():
            logger.debug("Connection to %s already in progress. Waiting first.", self.data.name)

        if self._client is not None and self._client.is_connected is True:
            return

        async with self._connect_lock:
            # Also check after lock is acquired
            if self._client is not None and self._client.is_connected is True:
                return
            if self._client is not None:
                with contextlib.suppress(BleakError, TimeoutError):
                    logger.debug("client exists, but not connected. Trying to connect.")
                    await self._client.connect()
                if self._client.is_connected is True:
                    logger.debug("Reconnect successful.")
                    return
            try:
                logger.debug("Establishing a new connection from mug (ID: %s) to %s", id(self), self.device)
                self._client = await establish_connection(
                    client_class=BleakClient,
                    device=self.device,
                    name=f'{self.data.name} ({self.device.address})',
                    use_services_cache=True,
                    disconnected_callback=self._disconnect_callback,  # type: ignore
                    ble_device_callback=lambda: self.device,
                )
                self._expected_disconnect = False
            except (asyncio.TimeoutError, BleakError) as error:
                logger.error("%s: Failed to connect to the mug: %s", self.device, error)
                raise error
            # Attempt to pair for good measure
            try:
                await self._client.pair()
            except (BleakError, EOFError):
                pass
            except NotImplementedError:
                # workaround for Home Assistant ESPHome Proxy backend which does not allow pairing.
                logger.warning(
                    'Pairing not implemented. '
                    'If your mug is still in pairing mode (blinking blue) tap the button on the bottom to exit.',
                )
            await self.subscribe()
            await close_stale_connections(self.device)

    async def _read(self, characteristic: MugCharacteristic) -> bytearray:
        """Helper to read characteristic from Mug."""
        if self._operation_lock.locked():
            logger.debug("Operation already in progress. waiting for it to complete")
        async with self._operation_lock:
            data = await self._client.read_gatt_char(characteristic.uuid)
            logger.debug("Read attribute '%s' with value '%s'", characteristic, data)
            return data

    async def _write(self, characteristic: MugCharacteristic, data: bytearray) -> None:
        """Helper to write characteristic to Mug."""
        if self._operation_lock.locked():
            logger.debug("Operation already in progress. Waiting for it to complete")
        async with self._operation_lock:
            await self._ensure_connection()
            try:
                await self._client.write_gatt_char(characteristic.uuid, data)
                logger.debug("Wrote '%s' to attribute '%s'", data, characteristic)
            except BleakError as e:
                logger.error("Failed to write '%s' to attribute '%s': %s", data, characteristic, e)
                raise

    async def disconnect(self) -> None:
        """Disconnect from mug and stop listening to notifications."""
        logger.debug("Disconnect called")
        self._expected_disconnect = True
        if self._client and self._client.is_connected is True:
            async with self._connect_lock:
                await self.unsubscribe()
                await self._client.disconnect()
        self._expected_disconnect = False

    def _disconnect_callback(self, client: BleakClient) -> None:
        """Disconnect from device."""
        if self._expected_disconnect:
            logger.debug("Disconnect callback called")
        else:
            logger.warning("Unexpectedly disconnected")

    def _fire_callbacks(self) -> None:
        """Fire the callbacks."""
        logger.debug("Firing callbacks: %s", self._callbacks)
        for callback in self._callbacks:
            callback(self.data)

    def register_callback(self, callback: Callable[[MugData], None]) -> Callable[[], None]:
        """Register a callback to be called when the state changes."""
        if existing_unregister_callback := self._callbacks.get(callback):
            logger.debug("Callback %s already registered", callback)
            return existing_unregister_callback

        def unregister_callback() -> None:
            if callback in self._callbacks:
                del self._callbacks[callback]
            logger.debug("Unregistered callback: %s", callback)

        self._callbacks[callback] = unregister_callback
        logger.debug("Registered callback: %s", callback)
        return unregister_callback

    async def get_meta(self) -> MugMeta:
        """Fetch Meta info from the mug (Serial number and ID)."""
        return MugMeta.from_bytes(await self._read(MugCharacteristic.MUG_ID))

    async def get_battery(self) -> BatteryInfo:
        """Get Battery percent from mug gatt."""
        return BatteryInfo.from_bytes(await self._read(MugCharacteristic.BATTERY))

    async def get_led_colour(self) -> Colour:
        """Get RGBA colours from mug gatt."""
        return Colour.from_bytes(await self._read(MugCharacteristic.LED))

    async def set_led_colour(self, colour: Colour) -> None:
        """Set new target temp for mug."""
        colour = Colour(*colour[:3], 255)  # It always expects 255 for alpha
        await self._write(MugCharacteristic.LED, colour.as_bytearray())
        self.data.led_colour = colour

    async def get_target_temp(self) -> float:
        """Get target temp form mug gatt."""
        temp_bytes = await self._read(MugCharacteristic.TARGET_TEMPERATURE)
        return temp_from_bytes(temp_bytes, self.data.use_metric)

    async def set_target_temp(self, target_temp: float) -> None:
        """Set new target temp for mug."""
        target = bytearray(int(target_temp / 0.01).to_bytes(2, "little"))
        await self._write(MugCharacteristic.TARGET_TEMPERATURE, target)
        self.data.target_temp = target_temp

    async def get_current_temp(self) -> float:
        """Get current temp from mug gatt."""
        temp_bytes = await self._read(MugCharacteristic.CURRENT_TEMPERATURE)
        return temp_from_bytes(temp_bytes, self.data.use_metric)

    async def get_liquid_level(self) -> int:
        """Get liquid level from mug gatt."""
        liquid_level_bytes = await self._read(MugCharacteristic.LIQUID_LEVEL)
        return bytes_to_little_int(liquid_level_bytes)

    async def get_liquid_state(self) -> LiquidState:
        """Get liquid state from mug gatt."""
        liquid_state_bytes = await self._read(MugCharacteristic.LIQUID_STATE)
        state = bytes_to_little_int(liquid_state_bytes)
        return LiquidState(state)

    async def get_name(self) -> str:
        """Get mug name from gatt."""
        name_bytes: bytearray = await self._read(MugCharacteristic.MUG_NAME)
        return bytes(name_bytes).decode("utf8")

    async def set_name(self, name: str) -> None:
        """Assign new name to mug."""
        if MUG_NAME_REGEX.match(name) is None:
            raise ValueError('Name cannot contain any special characters')
        await self._write(MugCharacteristic.MUG_NAME, bytearray(name.encode("utf8")))
        self.data.name = name

    async def get_udsk(self) -> str:
        """Get mug udsk from gatt."""
        try:
            return decode_byte_string(await self._read(MugCharacteristic.UDSK))
        except BleakError as e:
            logger.debug('Unable to read UDSK: %s', e)
        return ''

    async def set_udsk(self, udsk: str) -> None:
        """Attempt to write udsk."""
        await self._write(MugCharacteristic.UDSK, bytearray(encode_byte_string(udsk)))
        self.data.udsk = udsk

    async def get_dsk(self) -> str:
        """Get mug dsk from gatt."""
        try:
            return decode_byte_string(await self._read(MugCharacteristic.DSK))
        except BleakError as e:
            logger.debug('Unable to read DSK: %s', e)
        return ''

    async def get_temperature_unit(self) -> TemperatureUnit:
        """Get mug temp unit."""
        unit_bytes = await self._read(MugCharacteristic.TEMPERATURE_UNIT)
        if bytes_to_little_int(unit_bytes) == 0:
            return TemperatureUnit.CELSIUS
        return TemperatureUnit.FAHRENHEIT

    async def set_temperature_unit(self, unit: Literal["°C", "°F"] | TemperatureUnit | Enum) -> None:
        """Set mug unit."""
        text_unit = unit.value if isinstance(unit, Enum) else unit
        unit_bytes = bytearray([1 if text_unit == TemperatureUnit.FAHRENHEIT else 0])
        await self._write(MugCharacteristic.TEMPERATURE_UNIT, unit_bytes)
        self.data.temperature_unit = TemperatureUnit(unit)

    async def ensure_correct_unit(self) -> None:
        """Set mug unit if it's not what we want."""
        desired = TemperatureUnit.CELSIUS if self.data.use_metric else TemperatureUnit.FAHRENHEIT
        if self.data.temperature_unit != desired:
            await self.set_temperature_unit(desired)

    async def get_battery_voltage(self) -> int:
        """Get voltage and charge time."""
        battery_voltage_bytes = await self._read(MugCharacteristic.CONTROL_REGISTER_DATA)
        return bytes_to_big_int(battery_voltage_bytes[:1])

    async def get_date_time_zone(self) -> datetime | None:
        """Get date and time zone."""
        date_time_zone_bytes = await self._read(MugCharacteristic.DATE_TIME_AND_ZONE)
        time_value = bytes_to_big_int(date_time_zone_bytes[:4])
        # offset = bytes_to_big_int(date_time_zone_bytes[4:])
        return datetime.fromtimestamp(time_value, timezone.utc) if time_value > 0 else None

    async def get_firmware(self) -> MugFirmwareInfo:
        """Get firmware info."""
        return MugFirmwareInfo.from_bytes(await self._read(MugCharacteristic.FIRMWARE))

    async def update_initial(self) -> list[Change]:
        """Update attributes that don't normally change and don't need to be regularly updated."""
        return await self._update_multiple(self._initial_attrs)

    async def update_all(self) -> list[Change]:
        """Update all standard attributes."""
        return await self._update_multiple(self._update_attrs)

    async def _update_multiple(self, attrs: set[str]) -> list[Change]:
        """Helper to update a list of attributes from the mug."""
        logger.debug('Updating the following attributes: %s', attrs)
        await self._ensure_connection()
        changes = self.data.update_info(**{attr: await getattr(self, f"get_{attr}")() for attr in attrs})
        if changes:
            self._fire_callbacks()
        logger.debug('Attributes updated: %s', changes)
        return changes

    async def update_queued_attributes(self) -> list[Change]:
        """Update all attributes in queue."""
        logger.debug('Updating queued attributes: %s', self._queued_updates)
        if not self._queued_updates:
            return []
        queued_updates = set(self._queued_updates)
        self._queued_updates.clear()
        await self._ensure_connection()
        changes = self.data.update_info(**{attr: await getattr(self, f"get_{attr}")() for attr in queued_updates})
        if changes:
            self._fire_callbacks()
        return changes

    def _notify_callback(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
        """Push events from the mug to indicate changes."""
        event_id = data[0]
        now = time()
        if (last_time := self._latest_events.get(event_id)) and now - last_time < 5:
            return
        self._latest_events[event_id] = now

        logger.debug("Push event received from Mug (%s) - Data: %s.", event_id, data)

        # Check known IDs
        if event_id in PUSH_EVENT_BATTERY_IDS:
            # 1, 2 and 3 : Battery Change
            if event_id in (
                PushEvent.CHARGER_CONNECTED,
                PushEvent.CHARGER_DISCONNECTED,
            ):
                self.data.battery = BatteryInfo(
                    percent=self.data.battery.percent if self.data.battery else 0,
                    on_charging_base=event_id == PushEvent.CHARGER_CONNECTED,
                )
                self._fire_callbacks()
            # All indicate changes in battery
            self._queued_updates.add("battery")
        elif event_id == PushEvent.TARGET_TEMPERATURE_CHANGED:
            self._queued_updates.add("target_temp")
        elif event_id == PushEvent.DRINK_TEMPERATURE_CHANGED:
            self._queued_updates.add("current_temp")
        elif event_id == PushEvent.AUTH_INFO_NOT_FOUND:
            logger.warning("Auth info missing")
        elif event_id == PushEvent.LIQUID_LEVEL_CHANGED:
            self._queued_updates.add("liquid_level")
        elif event_id == PushEvent.LIQUID_STATE_CHANGED:
            self._queued_updates.add("liquid_state")
        elif event_id == PushEvent.BATTERY_VOLTAGE_STATE_CHANGED:
            self._queued_updates.add("battery_voltage")
        else:
            logger.debug('Unknown even received %s', event_id)

    async def unsubscribe(self) -> None:
        """Unsubscribe from Mug notifications."""
        logger.debug("Unsubscribe called")
        if not self._client:
            return
        with contextlib.suppress(BleakError):
            await self._client.stop_notify(MugCharacteristic.PUSH_EVENT.uuid)

    async def subscribe(self) -> None:
        """Subscribe to notifications from the mug."""
        try:
            logger.info("Subscribe to Push Events")
            await self._client.start_notify(MugCharacteristic.PUSH_EVENT.uuid, self._notify_callback)
        except Exception as e:
            logger.warning("Failed to subscribe to state attr: %s", e)

    def set_client_options(self, **kwargs: str) -> None:
        """Update options in case they need to overriden in some cases."""
        if kwargs.get('adapter') and IS_LINUX is False:
            raise ValueError('The adapter option is only valid for the Linux BlueZ Backend.')
        self._client_kwargs = {**kwargs}

    @contextlib.asynccontextmanager
    async def connection(self, **kwargs: str) -> AsyncIterator[EmberMug]:
        """Helper for establishing a connection and automatically closing it."""
        self.set_client_options(**kwargs)
        # This will happen automatically, but calling it now will give us immediate feedback
        await self._ensure_connection()
        yield self
        await self.disconnect()



"""Classes for representing data from the mug."""
from __future__ import annotations

from dataclasses import asdict, dataclass, is_dataclass
from typing import Any, NamedTuple

from .consts import ATTR_LABELS, EXTRA_ATTRS, LiquidState, TemperatureUnit
from .formatting import format_led_colour, format_liquid_level, format_temp
from .utils import bytes_to_little_int, decode_byte_string


class Change(NamedTuple):
    """Helper for storing changes to attributes."""

    attr: str
    old_value: Any
    new_value: Any

    def __str__(self) -> str:
        """Use str to format Change message."""
        return f'{self.attr.replace("_", " ").title()} changed from "{self.old_value}" to "{self.new_value}"'


class Colour(NamedTuple):
    """Simple helper for colour formatting."""

    red: int
    green: int
    blue: int
    alpha: int | None = None

    def as_hex(self) -> str:
        """Format colour array as hex string."""
        return '#' + ''.join(f'{c:02x}' for c in self if c is not None)

    def as_bytearray(self) -> bytearray:
        """Convert to byte array."""
        return bytearray(c for c in self if c is not None)

    @classmethod
    def from_bytes(cls, data: bytes) -> Colour:
        """Initialize from raw bytes."""
        return cls(*data[:3])

    def __str__(self) -> str:
        """For more useful cli output, format as hex."""
        return self.as_hex()


@dataclass
class BatteryInfo:
    """Battery Information."""

    percent: float
    on_charging_base: bool

    @classmethod
    def from_bytes(cls, data: bytes) -> BatteryInfo:
        """Initialize from raw bytes."""
        return cls(
            percent=round(float(data[0]), 2),
            on_charging_base=data[1] == 1,
        )

    def __str__(self) -> str:
        """String representation for printing."""
        return f'{self.percent}%, {"" if self.on_charging_base else "not "}on charging base'


@dataclass
class MugFirmwareInfo:
    """Firmware versions."""

    version: int
    hardware: int
    bootloader: int

    @classmethod
    def from_bytes(cls, data: bytes) -> MugFirmwareInfo:
        """Initialize from raw bytes."""
        return cls(
            version=bytes_to_little_int(data[:2]),
            hardware=bytes_to_little_int(data[2:4]),
            bootloader=bytes_to_little_int(data[4:]),
        )

    def __str__(self) -> str:
        """String representation for printing."""
        return ', '.join(
            (
                f'Version: {self.version}',
                f'Hardware: {self.hardware}',
                f'Bootloader: {self.bootloader}',
            ),
        )


@dataclass
class MugMeta:
    """Meta data for mug."""

    mug_id: str
    serial_number: str

    @classmethod
    def from_bytes(cls, data: bytes) -> MugMeta:
        """Initialize from raw bytes."""
        return cls(
            mug_id=decode_byte_string(data[:6]),
            serial_number=data[7:].decode("utf-8"),
        )

    def __str__(self) -> str:
        """String representation for printing."""
        return f'Mug ID: {self.mug_id}, Serial Number: {self.serial_number}'


@dataclass
class MugData:
    """Class to store/display the state of the mug."""

    # Options
    model: str
    use_metric: bool = True
    include_extra: bool = False

    # Attributes
    name: str = ""
    meta: MugMeta | None = None
    battery: BatteryInfo | None = None
    firmware: MugFirmwareInfo | None = None
    led_colour: Colour = Colour(255, 255, 255)
    liquid_state: LiquidState = LiquidState.UNKNOWN
    liquid_level: int = 0
    temperature_unit: TemperatureUnit = TemperatureUnit.CELSIUS
    current_temp: float = 0.0
    target_temp: float = 0.0
    dsk: str = ""
    udsk: str = ""
    date_time_zone: str = ""
    battery_voltage: str = ""

    @property
    def meta_display(self) -> str:
        """Return Meta infor based on preference."""
        if self.meta and not self.include_extra:
            return f'Serial Number: {self.meta.serial_number}'
        return str(self.meta)

    @property
    def led_colour_display(self) -> str:
        """Return colour as hex value."""
        return format_led_colour(self.led_colour)

    @property
    def liquid_state_display(self) -> str:
        """Human-readable liquid state."""
        return self.liquid_state.label

    @property
    def liquid_level_display(self) -> str:
        """Human-readable liquid level."""
        return format_liquid_level(self.liquid_level)

    @property
    def current_temp_display(self) -> str:
        """Human-readable current temp with unit."""
        return format_temp(self.current_temp, self.use_metric)

    @property
    def target_temp_display(self) -> str:
        """Human-readable target temp with unit."""
        return format_temp(self.target_temp, self.use_metric)

    def update_info(self, **kwargs: Any) -> list[Change]:
        """Update attributes of the mug if they haven't changed."""
        changes: list[Change] = []
        for attr, new_value in kwargs.items():
            if (old_value := getattr(self, attr)) != new_value:
                setattr(self, attr, new_value)
                changes.append(Change(attr, old_value, new_value))
        return changes

    def get_formatted_attr(self, attr: str) -> str | None:
        """Get the display value of a given attribute."""
        if display_value := getattr(self, f'{attr}_display', None):
            return display_value
        return getattr(self, attr)

    @property
    def formatted(self) -> dict[str, Any]:
        """Return human-readable names and values for all attributes for display."""
        return {
            label: self.get_formatted_attr(attr)
            for attr, label in ATTR_LABELS.items()
            if self.include_extra or attr not in EXTRA_ATTRS
        }

    def as_dict(self) -> dict[str, Any]:
        """Dump all attributes as dict for info/debugging."""
        data = {k: asdict(v) if is_dataclass(v) else v for k, v in asdict(self).items()}
        data.update(
            {
                f'{attr}_display': getattr(self, f'{attr}_display')
                for attr in ('led_colour', 'liquid_state', 'liquid_level', 'current_temp', 'target_temp')
            },
        )
        return data



"""CLI Interface."""
from __future__ import annotations

import asyncio
import contextlib
import platform
import re
import sys
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from typing import TYPE_CHECKING

from bleak import BleakError

from ..consts import ATTR_LABELS, EXTRA_ATTRS
from ..data import Colour
from ..mug import EmberMug
from ..scanner import discover_mugs, find_mug
from .helpers import CommandLoop, print_changes, print_info, print_table, validate_mac

if TYPE_CHECKING:
    from bleak.backends.device import BLEDevice

all_attrs = list(ATTR_LABELS) + list(EXTRA_ATTRS)
get_attribute_names = [n.replace('_', '-') for n in all_attrs]


async def get_mug(args: Namespace) -> EmberMug:
    """Helper to get the mug based on args."""
    device = await find_device(args)
    mug = EmberMug(device, use_metric=not args.imperial, include_extra=args.extra)
    if not args.raw:
        print('Connecting...')
    return mug


async def find_device(args: Namespace) -> BLEDevice:
    """Find a single device that has already been paired."""
    try:
        device = await find_mug(mac=args.mac, adapter=args.adapter)
    except BleakError as e:
        print(f'An error occurred trying to find a mug: {e}')
        sys.exit(1)
    if not device:
        print('No mug was found.')
        sys.exit(1)
    if not args.raw:
        print('Found mug:', device)
    return device


async def discover(args: Namespace) -> list[BLEDevice]:
    """Discover new devices in pairing mode."""
    try:
        mugs = await discover_mugs(mac=args.mac)
    except BleakError as e:
        print(f'An error occurred trying to discover mugs: {e}')
        sys.exit(1)
    if not mugs:
        print('No mugs were found. Be sure it is in pairing mode. Or use "find" if already paired.')
        sys.exit(1)

    for mug in mugs:
        if args.raw:
            print(mug.address)
        else:
            print('Found mug:', mug)
    return mugs


async def fetch_info(args: Namespace) -> None:
    """Fetch all information from a mug and end."""
    mug = await get_mug(args)
    async with mug.connection(adapter=args.adapter):
        if not args.raw:
            print('Connected.\nFetching Info')
        await mug.update_all()
    print_info(mug)


async def poll_mug(args: Namespace) -> None:
    """Fetch all information and keep polling for changes."""
    mug = await get_mug(args)
    async with mug.connection(adapter=args.adapter):
        if not args.raw:
            print('Connected.\nFetching Info')
        await mug.update_all()
        print_info(mug)
        if not args.raw:
            print('\nWatching for changes')
        for _ in CommandLoop():
            for _ in range(60):
                await asyncio.sleep(1)
                print_changes(await mug.update_queued_attributes(), mug.data.use_metric)
            # Every minute do a full update
            print_changes(await mug.update_all(), mug.data.use_metric)


async def get_mug_value(args: Namespace) -> None:
    """Get values from the mug and print them."""
    mug = await get_mug(args)
    data = {}
    attributes = [a.replace('-', '_') for a in args.attributes]
    async with mug.connection(adapter=args.adapter):
        for attr in attributes:
            value = await getattr(mug, f'get_{attr}')()
            setattr(mug.data, attr, value)
            data[attr] = value
    if args.raw:
        print('\n'.join(str(v) for v in data.values()))
    else:
        print_table([(ATTR_LABELS.get(attr, attr), str(mug.data.get_formatted_attr(attr))) for attr in data])


async def set_mug_value(args: Namespace) -> None:
    """Set one or more values on the mug."""
    attrs = ('name', 'target_temp', 'temperature_unit', 'led_colour')
    values = [(attr, value) for attr in attrs if (value := getattr(args, attr))]
    if not values:
        print('Please specify at least one attribute and value to set.')
        options = [f'--{a}' for a in attrs]
        print(f'Options: {", ".join(options)}')
        sys.exit(1)

    mug = await get_mug(args)
    async with mug.connection(adapter=args.adapter):
        for attr, value in values:
            method = getattr(mug, f'set_{attr.replace("-", "_")}')
            print(f'Setting {attr} to {value}')
            await method(value)


def colour_type(value: str) -> Colour:
    """Convert a hex or rgb colour to a Colour object."""
    print(value)
    if match := re.match(r'#?([0-9a-f]{6})', value, re.IGNORECASE):
        colour = match.group(1)
        return Colour(*tuple(int(colour[i : i + 2], 16) for i in (0, 2, 4)))

    with contextlib.suppress(ValueError, AssertionError):
        colours = [int(v) for v in value.split(',')]
        assert len(colours) == 3 and all(0 <= c <= 255 for c in colours)
        return Colour(*colours)

    raise ArgumentTypeError(f'"{value}" is not a valid rgba or hex colour')


class EmberMugCli:
    """Very simple CLI Interface to interact with a mug."""

    _commands = {
        'find': find_device,
        'discover': discover,
        'info': fetch_info,
        'poll': poll_mug,
        'get': get_mug_value,
        'set': set_mug_value,
    }

    def __init__(self) -> None:
        """Create parsers."""
        self.parser = ArgumentParser(prog='ember-mug', description='CLI to interact with an Ember Mug')
        shared_parser = ArgumentParser(add_help=False)
        shared_parser.add_argument(
            '-m',
            '--mac',
            action='store',
            type=validate_mac,
            help='Only look for this specific address',
        )
        shared_parser.add_argument('-r', '--raw', help='No formatting. One value per line.', action='store_true')
        if platform.system() == 'Linux':
            # Only works on Linux with BlueZ so don't add for others.
            shared_parser.add_argument(
                '-a',
                '--adapter',
                action='store',
                help='Use this Bluetooth adapter instead of the default one (for Bluez)',
            )
        subparsers = self.parser.add_subparsers(dest='command', required=True)
        subparsers.add_parser('find', description='Find the first paired mug', parents=[shared_parser])
        subparsers.add_parser('discover', description='Discover Mugs in pairing mode', parents=[shared_parser])
        info_parsers = ArgumentParser(add_help=False)
        info_parsers.add_argument('-e', '--extra', help='Show extra info', action='store_true')
        info_parsers.add_argument('--imperial', help='Use Imperial units', action='store_true')
        subparsers.add_parser('info', description='Fetch all info from mug', parents=[shared_parser, info_parsers])
        subparsers.add_parser('poll', description='Poll mug for information', parents=[shared_parser, info_parsers])
        get_parser = subparsers.add_parser('get', description='Get mug value', parents=[shared_parser, info_parsers])
        get_parser.add_argument(dest='attributes', metavar='ATTRIBUTE', choices=get_attribute_names, nargs='+')
        set_parser = subparsers.add_parser('set', description='Set mug value', parents=[shared_parser, info_parsers])
        set_parser.add_argument('--name', help='Name', required=False)
        set_parser.add_argument('--target-temp', help='Target Temperature', type=float, required=False)
        set_parser.add_argument('--temperature-unit', help='Temperature Unit', choices=['C', 'F'], required=False)
        set_parser.add_argument('--led-colour', help='LED Colour', type=colour_type, required=False)

    async def run(self) -> None:
        """Run the specified command based on subparser."""
        args = self.parser.parse_args()
        await self._commands[args.command](args)



"""Bluetooth UUIDs and other constants used for communicating with the mug."""
from __future__ import annotations

import re
from enum import Enum, IntEnum
from functools import cached_property
from typing import Literal
from uuid import UUID

# Bluetooth names of supported mugs
EMBER_BLUETOOTH_NAMES: tuple[str, ...] = ("Ember Ceramic Mug",)

# Format for all the mug's Bluetooth UUIDs
UUID_TEMPLATE = "fc54{:0>4x}-236c-4c94-8fa9-944a3e5353fa"


class TemperatureUnit(str, Enum):
    """Temperature Units."""

    CELSIUS: Literal["°C"] = "°C"
    FAHRENHEIT: Literal["°F"] = "°F"


class MugCharacteristic(IntEnum):
    """Characteristic IDs for the Mug."""

    # Name of mug in byte string (Read/Write)
    MUG_NAME = 1
    # Current Mug Temp (Read)
    CURRENT_TEMPERATURE = 2
    # Target Mug Temp (Read/Write)
    TARGET_TEMPERATURE = 3
    # Unit (0 -> Celsius, 1 -> Fahrenheit) (Read/Write)
    TEMPERATURE_UNIT = 4
    # Level (Between 0 -> 30 ?) 30 100% ?
    LIQUID_LEVEL = 5
    # Date/Time (Read/Write)
    DATE_TIME_AND_ZONE = 6
    # Battery Info (Read)
    BATTERY = 7
    # Integer representing what it is doing with the liquid (Read)
    LIQUID_STATE = 8
    # Volume - I think for the thermos
    VOLUME = 9
    # Last location - (Write)
    LAST_LOCATION = 10
    # Unsure what it does
    UUID_ACCELERATION = 11
    # Firmware info (Read)
    FIRMWARE = 12
    # [Unique ID]-[serial number] (Read)
    MUG_ID = 13
    # DSK - Unique ID used for auth in app (Read)
    DSK = 14
    # UDSK - Used for auth in app (Read/Write)
    UDSK = 15
    # int/temp lock - Address (Read/Write)
    CONTROL_REGISTER_ADDRESS = 16
    # Battery charge info (Read/Write)
    CONTROL_REGISTER_DATA = 17
    # To watch for changes from mug (Notify/Read)
    PUSH_EVENT = 18
    # To gather bytes from mug for stats (Notify)
    STATISTICS = 19
    # RGBA Colour of LED (Read/Write)
    LED = 20
    # Service
    SERVICE = 13858

    @cached_property
    def uuid(self) -> UUID:
        """Convert the ID to a full UUID and cache."""
        return UUID(UUID_TEMPLATE.format(self.value))

    def __str__(self) -> str:
        """String representation is the UUID."""
        return str(self.uuid)


class LiquidState(IntEnum):
    """Constants for liquid state codes."""

    UNKNOWN = 0
    EMPTY = 1
    FILLING = 2
    COLD_NO_TEMP_CONTROL = 3
    COOLING = 4
    HEATING = 5
    TARGET_TEMPERATURE = 6
    WARM_NO_TEMP_CONTROL = 7

    @cached_property
    def label(self) -> str:
        """Get label for current state."""
        return LIQUID_STATE_LABELS[self.value]

    def __str__(self) -> str:
        """String is the label."""
        return self.label


# Push event codes
class PushEvent(IntEnum):
    """IDs for Push Events."""

    BATTERY_CHANGED = 1
    CHARGER_CONNECTED = 2
    CHARGER_DISCONNECTED = 3
    TARGET_TEMPERATURE_CHANGED = 4
    DRINK_TEMPERATURE_CHANGED = 5
    AUTH_INFO_NOT_FOUND = 6
    LIQUID_LEVEL_CHANGED = 7
    LIQUID_STATE_CHANGED = 8
    BATTERY_VOLTAGE_STATE_CHANGED = 9


# Labels so liquid states
LIQUID_STATE_LABELS: dict[int, str] = {
    LiquidState.UNKNOWN: "Unknown",
    LiquidState.EMPTY: "Empty",
    LiquidState.FILLING: "Filling",
    LiquidState.COLD_NO_TEMP_CONTROL: "Cold (No control)",
    LiquidState.COOLING: "Cooling",
    LiquidState.HEATING: "Heating",
    LiquidState.TARGET_TEMPERATURE: "Perfect",
    LiquidState.WARM_NO_TEMP_CONTROL: "Warm (No control)",
}

PUSH_EVENT_BATTERY_IDS = [
    PushEvent.BATTERY_CHANGED,
    PushEvent.CHARGER_CONNECTED,
    PushEvent.CHARGER_DISCONNECTED,
]

# Labels for formatting attributes
ATTR_LABELS = {
    'name': 'Mug Name',
    'meta': 'Meta',
    'battery': 'Battery',
    'firmware': 'Firmware',
    'led_colour': 'LED Colour',
    'liquid_state': 'Liquid State',
    'liquid_level': 'Liquid Level',
    'current_temp': 'Current Temp',
    'target_temp': 'Target Temp',
    'use_metric': 'Use Metric',
    'dsk': 'DSK',
    'udsk': 'UDSK',
    'date_time_zone': 'Date Time + Time Zone',
    'battery_voltage': 'Voltage',
}

# Attributes
INITIAL_ATTRS = {
    "meta",
    "udsk",
    "dsk",
    "date_time_zone",
    "firmware",
}
UPDATE_ATTRS = {
    "name",
    "led_colour",
    "current_temp",
    "target_temp",
    "temperature_unit",
    "battery",
    "liquid_level",
    "liquid_state",
    "battery_voltage",
}
EXTRA_ATTRS = {'dsk', 'udsk', 'battery_voltage', 'date_time_zone'}

# Validation
MUG_NAME_REGEX = re.compile(r"^[A-Za-z0-9,.\[\]#()!\"\';:|\-_+<>%= ]{1,16}$")
MAC_ADDRESS_REGEX = re.compile(r"^([0-9A-Fa-f]{2}:){5}([0-9A-Fa-f]{2})$")



from __future__ import annotations

from argparse import Namespace
from collections.abc import Generator
from unittest.mock import AsyncMock, MagicMock, Mock, patch

import pytest
from bleak import BleakError, BLEDevice
from pytest import CaptureFixture

from ember_mug import EmberMug
from ember_mug.cli.commands import EmberMugCli, discover, fetch_info, find_device, get_mug, get_mug_value
from ember_mug.data import MugData
from tests.conftest import TEST_MAC

from ..conftest import mock_connection


@pytest.fixture()
def mock_mug_with_connection() -> Generator[AsyncMock, None, None]:
    with patch('ember_mug.cli.commands.get_mug') as mock:
        mock_mug = AsyncMock()
        mock_mug.connection = Mock(return_value=mock_connection)
        mock.return_value = mock_mug
        yield mock_mug


@patch('ember_mug.cli.commands.EmberMug', spec=EmberMug)
@patch('ember_mug.cli.commands.find_device')
async def test_get_mug(
    mock_find_device: AsyncMock,
    mock_ember_mug: AsyncMock,
    capsys: CaptureFixture,
    ble_device: BLEDevice,
) -> None:
    mock_find_device.return_value = ble_device
    args = Namespace(imperial=False, extra=True, raw=False)
    mug = await get_mug(args)
    assert mug is not None
    mock_find_device.assert_called_once_with(args)
    mock_ember_mug.assert_called_once_with(ble_device, use_metric=True, include_extra=True)
    captured = capsys.readouterr()
    assert captured.out == "Connecting...\n"

    # Raw prints nothing
    args = Namespace(imperial=False, extra=True, raw=True)
    mug = await get_mug(args)
    assert mug is not None
    captured = capsys.readouterr()
    assert captured.out == ""


@patch('ember_mug.cli.commands.find_mug')
async def test_find_device(mock_find_mug: AsyncMock, capsys: CaptureFixture, ble_device: BLEDevice) -> None:
    mock_find_mug.return_value = ble_device
    args = Namespace(mac=ble_device.address, adapter=None, raw=False)
    device = await find_device(args)
    assert device == ble_device
    mock_find_mug.assert_called_once_with(mac=ble_device.address, adapter=None)
    captured = capsys.readouterr()
    assert captured.out == f"Found mug: {ble_device}\n"

    # Raw prints nothing
    args = Namespace(mac=ble_device.address, adapter=None, raw=True)
    await find_device(args)
    captured = capsys.readouterr()
    assert captured.out == f""


@patch('ember_mug.cli.commands.find_mug')
async def test_find_device_no_device(mock_find_mug: AsyncMock, capsys: CaptureFixture) -> None:
    mock_find_mug.return_value = None
    args = Namespace(mac=TEST_MAC, adapter=None, raw=False)
    with pytest.raises(SystemExit):
        await find_device(args)
    mock_find_mug.assert_called_once_with(mac=TEST_MAC, adapter=None)
    captured = capsys.readouterr()
    assert captured.out == "No mug was found.\n"


@patch('ember_mug.cli.commands.find_mug')
async def test_find_device_bleak_error(mock_find_mug: AsyncMock, capsys: CaptureFixture) -> None:
    mock_find_mug.side_effect = BleakError('Test Error')
    args = Namespace(mac=TEST_MAC, adapter=None, raw=False)
    with pytest.raises(SystemExit):
        await find_device(args)
    mock_find_mug.assert_called_once_with(mac=TEST_MAC, adapter=None)
    captured = capsys.readouterr()
    assert captured.out == "An error occurred trying to find a mug: Test Error\n"


@patch('ember_mug.cli.commands.discover_mugs')
async def test_discover(mock_discover_mugs: AsyncMock, capsys: CaptureFixture, ble_device: BLEDevice) -> None:
    mock_discover_mugs.return_value = [ble_device]
    args = Namespace(mac=TEST_MAC, adapter=None, raw=False)
    mugs = await discover(args)
    assert mugs == [ble_device]
    mock_discover_mugs.assert_called_once_with(mac=TEST_MAC)
    captured = capsys.readouterr()
    assert captured.out == f"Found mug: {ble_device}\n"

    mock_discover_mugs.reset_mock()
    args = Namespace(mac=TEST_MAC, adapter=None, raw=True)
    mugs = await discover(args)
    assert mugs == [ble_device]
    mock_discover_mugs.assert_called_once_with(mac=TEST_MAC)
    captured = capsys.readouterr()
    assert captured.out == f"{TEST_MAC}\n"


@patch('ember_mug.cli.commands.discover_mugs')
async def test_discover_no_device(mock_discover_mugs: AsyncMock, capsys: CaptureFixture) -> None:
    mock_discover_mugs.return_value = []
    args = Namespace(mac=TEST_MAC, adapter=None, raw=False)
    with pytest.raises(SystemExit):
        await discover(args)
    mock_discover_mugs.assert_called_once_with(mac=TEST_MAC)
    captured = capsys.readouterr()
    assert captured.out == "No mugs were found. Be sure it is in pairing mode. Or use \"find\" if already paired.\n"


@patch('ember_mug.cli.commands.discover_mugs')
async def test_discover_bleak_error(mock_discover_mugs: AsyncMock, capsys: CaptureFixture) -> None:
    mock_discover_mugs.side_effect = BleakError('Test Error')
    args = Namespace(mac=TEST_MAC, adapter=None, raw=False)
    with pytest.raises(SystemExit):
        await discover(args)
    mock_discover_mugs.assert_called_once_with(mac=TEST_MAC)
    captured = capsys.readouterr()
    assert captured.out == "An error occurred trying to discover mugs: Test Error\n"


@patch('ember_mug.cli.commands.print_info')
async def test_fetch_info(
    mock_print_info: AsyncMock,
    mock_mug_with_connection: AsyncMock,
    capsys: CaptureFixture,
) -> None:
    # Test normal
    args = Namespace(mac=TEST_MAC, adapter=None, raw=False)
    await fetch_info(args)
    captured = capsys.readouterr()
    assert captured.out == "Connected.\nFetching Info\n"
    mock_print_info.assert_called_once_with(mock_mug_with_connection)

    # Test with Raw
    args = Namespace(mac=TEST_MAC, adapter=None, raw=True)
    await fetch_info(args)
    captured = capsys.readouterr()
    assert captured.out == ""


@patch('ember_mug.cli.commands.print_table')
async def test_get_mug_value(
    mocked_print_table: Mock,
    mock_mug_with_connection: AsyncMock,
    capsys: CaptureFixture,
) -> None:
    mock_mug_with_connection.data = MugData('test')
    mock_mug_with_connection.data.get_formatted_attr = Mock(return_value='test')  # type: ignore[assignment]
    mock_mug_with_connection.get_target_temp.return_value = 55.5
    mock_mug_with_connection.get_name.return_value = 'test'
    args = Namespace(adapter=None, raw=False, attributes=['target_temp', 'name'])
    await get_mug_value(args)
    mock_mug_with_connection.get_target_temp.assert_called_once()
    mocked_print_table.assert_called_once_with([('Target Temp', "test"), ('Mug Name', "test")])

    mock_mug_with_connection.get_led_colour.return_value = 55.5
    args = Namespace(adapter=None, raw=True, attributes=['led_colour', 'name'])
    await get_mug_value(args)
    captured = capsys.readouterr()
    assert captured.out == "55.5\ntest\n"


def test_ember_cli():
    cli = EmberMugCli()
    args = cli.parser.parse_args(['find'])
    assert args.command == 'find'

    args = cli.parser.parse_args(['discover'])
    assert args.command == 'discover'

    args = cli.parser.parse_args(['info', '-m', TEST_MAC, '--imperial'])
    assert args.command == 'info'
    assert args.mac == TEST_MAC
    assert args.imperial is True

    args = cli.parser.parse_args(['poll'])
    assert args.command == 'poll'

    args = cli.parser.parse_args(['get', 'led-colour'])
    assert args.command == 'get'
    assert args.attributes == ['led-colour']

    args = cli.parser.parse_args(['set', '--name', 'TEST'])
    assert args.command == 'set'
    assert args.name == 'TEST'


@patch('sys.argv', ['file.py', 'find', '-m', TEST_MAC])
async def test_cli_run():
    cli = EmberMugCli()
    mock_find = AsyncMock()
    with patch.object(cli, '_commands', {'find': mock_find}):
        await cli.run()
    mock_find.assert_called_once()



"""Tests for `ember_mug.mug`."""
from __future__ import annotations

from typing import Any

from ember_mug.consts import LiquidState, TemperatureUnit
from ember_mug.data import BatteryInfo, Change, Colour, MugData, MugFirmwareInfo, MugMeta


def test_mug_formatting(mug_data: MugData) -> None:
    meta = MugMeta('A', 'ABCDEF')
    battery = BatteryInfo(35.46, False)
    firmware = MugFirmwareInfo(12, 35, 55)
    colour = Colour(244, 0, 161)
    mug_data.update_info(
        name='Mug Name',
        meta=meta,
        battery=battery,
        firmware=firmware,
        led_colour=colour,
        liquid_state=LiquidState.HEATING,
        liquid_level=5,
        current_temp=25,
        target_temp=55,
        dsk='dsk',
        udsk='udsk',
        date_time_zone=None,
        battery_voltage=0.1,
    )
    assert mug_data.meta_display == "Serial Number: ABCDEF"
    mug_data.include_extra = True
    assert mug_data.meta_display == "Mug ID: A, Serial Number: ABCDEF"
    assert mug_data.led_colour_display == "#f400a1"
    assert mug_data.liquid_state_display == "Heating"
    assert mug_data.liquid_level_display == "16.67%"
    assert mug_data.current_temp_display == "25.00°C"
    assert mug_data.target_temp_display == "55.00°C"
    basic_info: dict[str, Any] = {
        'Mug Name': 'Mug Name',
        'Meta': "Serial Number: ABCDEF",
        'Battery': battery,
        'Firmware': firmware,
        'LED Colour': "#f400a1",
        'Liquid State': "Heating",
        'Liquid Level': "16.67%",
        'Current Temp': "25.00°C",
        'Target Temp': "55.00°C",
        'Use Metric': True,
    }
    assert mug_data.formatted == {
        **basic_info,
        'Meta': "Mug ID: A, Serial Number: ABCDEF",
        'DSK': "dsk",
        'UDSK': "udsk",
        'Date Time + Time Zone': None,
        'Voltage': 0.1,
    }
    mug_data.include_extra = False
    assert mug_data.formatted == basic_info


def test_update_info(mug_data: MugData) -> None:
    mug_data.current_temp = 55
    mug_data.target_temp = 55
    mug_data.led_colour = Colour(255, 0, 0)
    changes = mug_data.update_info(
        current_temp=55,
        target_temp=68,
        led_colour=Colour(0, 255, 0),
    )
    assert changes == [
        Change('target_temp', 55, 68),
        Change('led_colour', Colour(255, 0, 0), Colour(0, 255, 0)),
    ]


def test_mug_dict(mug_data: MugData) -> None:
    mug_data.update_info(meta=MugMeta('test_id', 'serial number'))
    assert mug_data.as_dict() == {
        'model': 'Ember Ceramic Mug',
        'include_extra': False,
        'use_metric': True,
        'battery': None,
        'battery_voltage': '',
        'current_temp': 0.0,
        'current_temp_display': '0.00°C',
        'date_time_zone': '',
        'dsk': '',
        'firmware': None,
        'led_colour': Colour(red=255, green=255, blue=255),
        'led_colour_display': '#ffffff',
        'liquid_level': 0,
        'liquid_level_display': '0.00%',
        'liquid_state': LiquidState.UNKNOWN,
        'liquid_state_display': 'Unknown',
        'meta': {'mug_id': 'test_id', 'serial_number': 'serial number'},
        'name': '',
        'target_temp': 0.0,
        'target_temp_display': '0.00°C',
        'temperature_unit': TemperatureUnit.CELSIUS,
        'udsk': '',
    }



from __future__ import annotations

from argparse import ArgumentTypeError
from textwrap import dedent

import pytest
from pytest import CaptureFixture

from ember_mug import EmberMug
from ember_mug.cli.helpers import build_sub_rows, print_changes, print_info, print_table, validate_mac
from ember_mug.consts import LiquidState
from ember_mug.data import Change


def test_validate_mac() -> None:
    with pytest.raises(ArgumentTypeError):
        validate_mac('potato')
    assert validate_mac('9C:DA:8C:19:27:DA') == '9c:da:8c:19:27:da'


def test_build_sub_rows() -> None:
    sub_rows = build_sub_rows(('Test', 'test1, test2, test3', 'test4'))
    assert sub_rows[0][0] == 'Test'
    assert sub_rows[0][1] == 'test1'
    assert sub_rows[0][2] == 'test4'
    assert sub_rows[1][0] == ''
    assert sub_rows[1][1] == 'test2'
    assert sub_rows[1][2] == ''
    assert sub_rows[2][1] == 'test3'


def test_print_changes(capsys: CaptureFixture) -> None:
    changes = [
        Change('name', 'Mug Name', 'Test Mug'),
        Change('liquid_level', 1, 2),
        Change('liquid_state', LiquidState.EMPTY, LiquidState.HEATING),
        Change('target_temp', 45, 55),
    ]
    print_changes(changes, True)
    captured = capsys.readouterr()
    assert captured.out == dedent(
        """\
        Name changed from "Mug Name" to "Test Mug"
        Liquid Level changed from "3.33%" to "6.67%"
        Liquid State changed from "Empty" to "Heating"
        Target Temp changed from "45.00°C" to "55.00°C"
        """,
    )


def test_print_table(ember_mug: EmberMug, capsys: CaptureFixture) -> None:
    print_table([])
    captured = capsys.readouterr()
    assert captured.out == ''


def test_print_info(ember_mug: EmberMug, capsys: CaptureFixture) -> None:
    print_info(ember_mug)
    captured = capsys.readouterr()
    assert captured.out == dedent(
        """\
        Mug Data
        +--------------+---------+
        | Mug Name     |         |
        +--------------+---------+
        | Meta         | None    |
        +--------------+---------+
        | Battery      | None    |
        +--------------+---------+
        | Firmware     | None    |
        +--------------+---------+
        | LED Colour   | #ffffff |
        +--------------+---------+
        | Liquid State | Unknown |
        +--------------+---------+
        | Liquid Level | 0.00%   |
        +--------------+---------+
        | Current Temp | 0.00°C  |
        +--------------+---------+
        | Target Temp  | 0.00°C  |
        +--------------+---------+
        | Use Metric   | True    |
        +--------------+---------+
        """,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ember_mug', 'ember_mug.cli', 'tests', 'tests.cli']

package_data = \
{'': ['*']}

install_requires = \
['bleak-retry-connector>=2.13.0', 'bleak>=0.19.5']

entry_points = \
{'console_scripts': ['ember-mug = ember_mug.cli:run_cli']}

setup_kwargs = {
    'name': 'python-ember-mug',
    'version': '0.6.0b6',
    'description': 'Python Library for Ember Mugs.',
    'long_description': '# Python Ember Mug\n\n[![pypi](https://img.shields.io/pypi/v/python-ember-mug.svg)](https://pypi.org/project/python-ember-mug/)\n[![python](https://img.shields.io/pypi/pyversions/python-ember-mug.svg)](https://pypi.org/project/python-ember-mug/)\n[![Build Status](https://github.com/sopelj/python-ember-mug/actions/workflows/dev.yml/badge.svg)](https://github.com/sopelj/python-ember-mug/actions/workflows/dev.yml)\n[![codecov](https://codecov.io/gh/sopelj/python-ember-mug/branch/main/graphs/badge.svg)](https://codecov.io/github/sopelj/python-ember-mug)\n![Project Maintenance](https://img.shields.io/maintenance/yes/2023.svg?style=for-the-badge)\n\nPython Library for interacting with Ember Mugs over Bluetooth\n\n* Documentation: <https://sopelj.github.io/python-ember-mug>\n* GitHub: <https://github.com/sopelj/python-ember-mug>\n* PyPI: <https://pypi.org/project/python-ember-mug/>\n* Free software: MIT\n\n## Summary\n\nThis is an *unofficial* library to attempt to interact with Ember Mugs via Bluetooth.\nThis was created for use with my [Home Assistant integration](https://github.com/sopelj/hass-ember-mug-component),\nbut could be useful separately and has a simple CLI interface too.\n\n**Note**: Should work with the standard Ember Mugs (1 and 2), but the thermoses have not been tested.\n\n## Features\n\n* Finding mugs\n* Connecting to Mugs\n* Reading Information (Colour, temp, liquid level, etc.)\n* Writing (Desired temp, colour, temperature unit)*\n* Polling for changes\n\n*** Writing only works if the mug has been set up in the app previously\n\n## Usage\n\n### Python\n\n```python\nfrom ember_mug.scanner import find_mug, discover_mugs\nfrom ember_mug.mug import EmberMug\n\n# if first time with mug in pairing\nmugs = await discover_mugs()\ndevice = mugs[0]\n# after paired you can simply use\ndevice = await find_mug()\nmug = EmberMug(device)\nasync with mug.connection() as con:\n    print(\'Connected.\\nFetching Info\')\n    await con.update_all()\n    print(mug.formatted_data)\n```\n\n### CLI\n\nIt can also be run via command line either directly with `ember-mug --help` or as a module with `python -m ember_mug --help`\nThere are four options with different subsections. You can see them by specifying them before help. eg `ember-mug poll --help`\n\n```bash\nember-mug discover  # Finds the mug in pairing mode for the first time\nember-mug poll  # fetches info and keeps listening for notifications\nember-mug get name target-temp  # Prints name and target temp of mug\nember-mug set --name "My mug" --target-temp 56.8  # Sets the name and target temp to specified values\n```\n\nBasic options:\n\n| Command     | Use                                                                            |\n|-------------|--------------------------------------------------------------------------------|\n| `discover`  | Find/List all detected unpaired mugs in pairing mode                           |\n| `find`      | Find *one* already paired mugs                                                 |\n| `info`      | Connect to *one* mug and print its current state                               |\n| `poll`      | Connect to *one* mug and print its current state and keep watching for changes |\n| `get`       | Get the value(s) of one or more attribute(s) by name                           |\n| `set`       | Set one or more values on the mug                                              |\n\n\n![CLI Example](./docs/images/cli-example.png)\n\n## Caveats\n\n- Since this api is not public, a lot of guesswork and reverse engineering is involved, so it\'s not perfect.\n- If the mug has not been set up in the app since it was reset, writing is not allowed. I don\'t know what they set in the app, but it changes something, and it doesn\'t work without it.\n- Once that mug has been set up in the app, you should ideally forget the device or at least turn off bluetooth whilst using it here, or you will probably get disconnected often\n- I haven\'t figured out some attributes like udsk, dsk, location and timezone.\n\n## Troubleshooting\n\n### \'Operation failed with ATT error: 0x0e\' or another connection error\nThis seems to be caused by the bluetooth adaptor being in some sort of passive mode. I have not yet figured out how to wake it programmatically so sadly, you need to manually open `bluetoothctl` to do so.\nPlease ensure the mug is in pairing mode (ie the light is flashing blue) and run the `bluetoothctl` command. You don,t need to type anything. run it and wait until the mug connects.\n\n## Todo\n- Test with other devices. Please let me know if you have tried it with others.\n\n## Credits\n\nThis package was created with [Cookiecutter](https://github.com/audreyr/cookiecutter) and the [waynerv/cookiecutter-pypackage](https://github.com/waynerv/cookiecutter-pypackage) project template.\n\n## Notice of Non-Affiliation and Disclaimer\n\nThis project is not affiliated, associated, authorized, endorsed by, or in any way officially connected with Ember.\n\nThe name Ember as well as related names, marks, emblems and images are registered trademarks of their respective owners.\n',
    'author': 'Jesse Sopel',
    'author_email': 'jesse.sopel@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sopelj/python-ember-mug',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



"""Tests for `ember_mug.consts`."""
from __future__ import annotations

from uuid import UUID

from ember_mug.consts import LiquidState, MugCharacteristic


def test_mug_uuids() -> None:
    assert MugCharacteristic.MUG_NAME.uuid == UUID("fc540001-236c-4c94-8fa9-944a3e5353fa")
    assert str(MugCharacteristic.MUG_NAME) == "fc540001-236c-4c94-8fa9-944a3e5353fa"


def test_liquid_state() -> None:
    assert LiquidState(0).label == "Unknown"
    assert LiquidState(1).label == "Empty"
    assert LiquidState(2).label == "Filling"
    assert LiquidState(3).label == "Cold (No control)"
    assert LiquidState(4).label == "Cooling"
    assert LiquidState(5).label == "Heating"
    assert LiquidState(6).label == "Perfect"
    assert LiquidState(7).label == "Warm (No control)"



"""Tests for `ember_mug.formatting`."""
from ember_mug.data import Colour
from ember_mug.formatting import format_led_colour, format_liquid_level, format_temp


def test_format_led_colour() -> None:
    assert format_led_colour(Colour(244, 0, 161)) == "#f400a1"


def test_format_liquid_level() -> None:
    assert format_liquid_level(5) == "16.67%"
    assert format_liquid_level(6) == "20.00%"


def test_format_temp() -> None:
    assert format_temp(25.445) == '25.45°C'
    assert format_temp(36.443, metric=False) == '36.44°F'
